
/* - get different expectation values for the eigenvector
 * - measure operators based on operators defined in hilbert.c
 * - operators can be defined by the user (see wop())

  ToDo:
     + sbase => hbuf???
     + convert old noSym functions to Operators for Opij-function
  v2.23 - add fermionic operators (2site, 4site), add comments
  
*/
#include "spins.h"
#include "xval.h"
#include "symmetry.h"
#include "tbase.h"
#include "block.h"
#include "hilbert.h"
#include "array_m.h"
#include "vector.h"

ULong	x_max;	/* maximum number of coeff. 0..MAXx set in daten.i */

/* overlap only works for the kagome 3003, 6022 and squago 2002 model */
int modeloverlap(tvec *v0){
#if 0 && CM(1,0,0)  /* only Heisenberg */
 if (nw==2*nn) /* Kagome */
 if (nn==24 || nn==27 || nn==36) {  /*  new algo, correct sign */
   int i1,i2,i3,i4,i5,i6,j1,j2,j3,j4,k3;
   ULong k,k2=0; double r1,r3,r4; cplx rr,r2,r5,ms;
   ULong n2,i5s,n3; tbase b1;
   static struct sbase b2; b2.o=0;
   if (mpi_n>1) E0("missing mpi code on modeloverlap");
#if 0  /* this function only works for the kagome 6 0 2 2 model */
   # space for 4 independent Magnons with different signs
   # plateaus: Sz=18-4(ea=8),14-4(ea=1),10-4=6
   #             -(35)-19--20--22--23--25--26--28--29--31--32--34--35--(19)
   #               /   ~\  /    \  /    \  /    \  /    \  /    \  /
   #            (33)  X  18      21      24  X   27      30      33
   #       \    /  \    /  \    /  \    / ~\    / ~\    /  \    /  \
   #  (17)--1---2---4---5---7---8--10--11--13--14--16--17-(1)-(2)-(4)-
   #         \ /    ~\ /     \ /    ~\ /     \ /     \ /    \ /    ~\
   #      X   0       3       6   O   9      12      15  X  (0)
   #         /~\     / \     /~\     / \     / \     /~\     ~
   #      (23)-(25)-26--28--29--31--32--34--35-(19)-20--22
   # PM600: v2.22cplx.i80=2m44s incv=3m45s a*b=7m37s
   #      sym= -6 0 0 -9000 (72sym)        a*b=9m17s
   #      no_sym n1=30e6 270MB(flt)-1GB(cplx)
   # Sz=10 nud=28+8 GZ  -4.9410286   4 q_0_0_0  m=5/9  n1=211396
   #   ----------------------------- sym= 144       72        1 ------
   #   n1                              211396   421200 30260340
   #   num_projected_sym_states          1122      ...       (1122*144=162e3)
   #   num_projected_states            151875       OK  = 15^4*3-27 ???
   #   weight_projected_states        0.54591719
   #   weight_projecting_states b*b   3.00313189 ?
   #   projection^2         (a*b)^2   1.57518385 ?
   #   overlap=sqrt((a*b/|b|)^2)      0.72423319    m=5/9 (proofed by Richter) q0-hex
   # max non_projected state  0.000017894 |uuuuuuuuduuuuduuuduuuuuduuduududuuud> (1
   #
   #   num_projected_states     480000 = 3*20^4
   #   weight_projected_states        0.41451702
   #   weight_projecting_states b*b   2.99449929 ? 
   #   projection^2         (a*b)^2   0.12730589 ?
   #   overlap=sqrt((a*b/|b|)^2)      0.20618741   m=1/3 E=-12.51761327 q3-hex
   #   max non_projected state  0.000056130 |uuduudduuduuuduuduuduuduuudduuduuuud> (1
   # ToDo: what about product states belonging to 2 different magnon crystals
   #  --------------------------------------------------------
   # til8_27 e=(3 0 0 3)
   #             6         0         3    *    6
   #              \       / \       / \       / 
   #          26---19---20---22---23---25---26---19
   #                 \ /       \ /       \ /
   #       24         18   *    21        24
   #         \       / \       / \       / \
   #     17---10---11---13---14---16---17---10
   #            \ /       \ /      \  /
   #   15   *    9        12        15    .
   #     \      / \       / \       / \
   # 8--- 1--- 2--- 4--- 5--- 7--- 8--- 1
   #       \ /       \ /       \ /
   #        0         3    .    6 
   #
   # num_projected_states          10125 = 15^3*3 (27 overlapping configs)
   # weight_projected_states        0.55482163
   # weight_projecting_states b*b   3.02392567 bit greater than 3
   # projection^2         (a*b)^2   1.50217824 
   # overlap=sqrt((a*b/|b|)^2)      0.70481506
   #  (Richter got the same result by diagonalizing overlap matrix for 3
   #   magnon basis variants, so we are sure now; the b*b>3 is caused by  
   #   overlap between 3 magnon basis variants (nonorthogonal base))
   #  |uuuuuuuuduudduuuuuuudduuduu>
   #  |.333.311.33..222.21.122..11> <-hexrings 1,2,3
   #  |3.311..33.222.233.22..111.1>
   #  |22..333.31.133..11.222.211.> 2*1d, 2*2d, 2*3d
   # -------------------------------------------------------------------
   # space for 4 independent Magnons with different signs
   #           4         10
   #          / \        / \
   #        15---14    21---20
   #     23< | * | >17< | * | >23
   #        12---13    18---19
   #          \ /        \ /
   #          16         22
   #          / \        / \
   #         3---2      9---8
   #     11< | * | > 5 <| * | >11
   #         0---1      6---7
   #          \ /        \ /
   #           4         10
   # Sz=8 nud=20+4 GZ E0=0  q_0_0  m=2/3  n1.sym=165 n1=10626
   #   ----------------------------- sym= 128       64        1 ------
   #  num_projected_states     256 = 4^4
   #  weight_projected_states        1.00000000
   #  weight_projecting_states b*b   1.00000000 (unsure)
   #  projection^2         (a*b)^2   1.00000000 (unsure)
   #  overlap=sqrt((a*b/|b|)^2)      1.00000000    ????
   # max non_projected state  0.000000000 |uuuuuuuuuuuuuuududuuudud> (1)
   # x_13 (weight, factor, coefficient, |sN...s0>, (symfactor)):
   #  0.250000000 1  0.500000000 |uuuuuduuuuuduuuuuduuuduu> (64)
   #  0.500000000 2 -0.500000000 |uuuuuduuuuuduuuuuduuuudu> (64)
   #  0.125000000 1  0.353553391 |uuuuuduuuuuduuuduuuuuduu> (32)
   #  0.125000000 2  0.250000000 |uuuuuduuuuuduuuuuduuuuud> (16)
   #                                ....  ....  ....  ....
   #  lc4:  E=-1 k=2- signum!
   # Sz=4 nud=16+8 GZ E0=-8.6454410  q_0_0_0_0_0+  m=1/3
   #   ----------------------------- sym= 128       32        1 ------
   # num_projected_states     1296 = 6^4
   # weight_projected_states        0.69167589    OK
   # weight_projecting_states b*b  16.00000000    OK (unsure) why 16?
   # projection^2         (a*b)^2  10.86312306    OK (unsure)
   # overlap=sqrt((a*b/|b|)^2)      0.82398131    OK (also for 6digits modelstate)
   #                                0.82403323   J1=1    J2=0.9999
   #                                0.99840967   J1=1    J2=0.1
   #                                0.99998342   J1=1    J2=0.01
   #                                1.00000000   J1=1    J2=1e-4
   #                                0.00003882   J1=1    J2=2
   # max non_projected state  0.000270717 |uuuuuduuudududududuuudud> (1)
   # x_6878 (weight, coefficient, |sN...s0>, (symfactor)):
   #  0.290790282     0.539249740 |uuuudduuududuuududuuudud> (128)
   #  0.097535081    -0.312306070 |uuududuuududuuududuuudud> (  8)
   #  0.097436300    -0.312147881 |uuududuuududuuududuududu> (  8)
   #  0.052928757    -0.230062507 |uuuudduuududuuudduuuudud> (128)
   #  0.034651819     0.186149991 |uuuuuduuudududududuuudud> (128)
   #                                 ....  ....  ....  ....
   #  0.573342238    eps: 1e-05 <= sum
   #  --------------------------------------------------------
   # x_2 (weight, coefficient, |sN...s0>, (symfactor)):  E=-2
   #  0.666666667  0.816496581 |udud> (2) ising.E1=-4 E2=4
   #  0.333333333 -0.408248290 |uudd> (2)       E1=0  E2=-4
   #  0.333333333 -0.408248290 |uddu> (2)       E1=0  E2=-4
#endif
   int vring=1, lring=4, nring=4; /* variants, length and num of rings */
   int ring4[]={ 0,1,2,3, 6,7,8,9, 12,13,14,15, 18,19,20,21, -1,-1,-1,-1};

   /* eigenstate of the N=4 ring, Sz=1 */
   int pstate4a[]={1,1,1,2, 1,1,2,1, 1,2,1,1, 2,1,1,1, -1,-1,-1,-1};
   double pa4a[]= {    0.5,    -0.5,     0.5,    -0.5,           0};

   /* eigenstate of the N=4 ring, Sz=0 */
#define  Q2d3  0.81649658092772603272 /* (2/3)^.5=0.81649658092772603272 */
#define  Q1d6  0.40824829046386301635 /* (1/6)^.5=0.40824829046386301635 */
   int pstate4b[]={1,2,1,2, 2,1,2,1, 1,1,2,2, 1,2,2,1, 2,2,1,1, 2,1,1,2, -1,-1,-1,-1};
   double pa4b[]= {Q2d3,Q2d3,-Q1d6,-Q1d6,-Q1d6,-Q1d6,0};

   /* we have 3 variants of 3 localized hexagons */
   int ring6_3[]={  /* pic2  8  3 0 0 3   til8_27 */
     0,1,8,6,19,20,  4,5,12,13,11,9, 16,17,24,25,23,21,
     0,22,23,3,4,2, 7,8,15,16,14,12, 10,11,18,19,26,24,
    1,2,9,10,17,15,   3,25,26,6,7,5, 13,14,21,22,20,18 };

   int ring6_4[]={  /* pic2  8  6 0 2 2  til8_36 */
    0,1,17,15,22,23, 4,5,18,19,35,33,    6,31,32,9,10,8, 13,14,27,28,26,24,
      0,25,26,3,4,2, 7,8,21,22,20,18,  9,34,35,12,13,11, 16,17,30,31,29,27,
    1,2,33,34,32,30,   3,28,29,6,7,5, 10,11,24,25,23,21, 12,19,20,15,16,14 };

   int ring6_5[]={  /* pic2  8  15 0 4 1 # til8_45 */
    0,1,44,42,13,14, 4,5,36,37,35,33, 9,10,8,6,22,23, 18,19,17,15,31,32, 27,28,26,24,40,41,
    3,4,2,0,16,17, 12,13,11,9,25,26, 21,22,20,18,34,35, 30,31,29,27,33,34, 39,40,38,36,42,43,
    6,7,5,3,19,20, 15,16,14,12,28,29, 24,25,23,21,37,38, 33,34,32,30,1,2, 42,43,41,39,10,11};

   int pstate6a[]={1,1,1,1,1,2,-1,-1,-1,-1,-1,-1}; /* space holder */
   double pa6a[]={0,0};
/* # x_3 (weight, coeff, |sN...s0>, (symfactor)): E=-2.11803399 q_0_0 m=5/9
   #  0.523606798  0.723606798 |uuudud> (6)  idx=1 E=-2 E2=+2  0.29541123825
   #  0.400000000 -0.632455532 |uuduud> (3)  idx=2 E=-2 E2=-2 -(2/15)^.5
   #  0.076393202 -0.276393202 |uuuudd> (6)  idx=0 E=+2 E2=-2 -0.11283705221
   q_3_1-
   #  1.000000000  1.000000000 |uuudud> (6)
   q_2-
   # E=-1.28077641 c2, Problem: complex + degeneracy
   # x_3 (weight, coefficient, diff, |sN...s0>, (symfactor)):
   #  0.526584766  0.725661605*exp(-0.000000*I*Pi) |uuudud> (6)
   #  0.431901719  0.657192300*exp(-0.666667*I*Pi) |uuduud> (3)
   #  0.041513516  0.203748658*exp(+0.666667*I*Pi) |uuuudd> (6)
 */

   int pstate6b[]={1,1,1,2,1,2, 1,1,2,1,2,1, 1,2,1,2,1,1, 2,1,2,1,1,1, 1,2,1,1,1,2, 2,1,1,1,2,1,
                   1,1,2,1,1,2, 1,2,1,1,2,1, 2,1,1,2,1,1,
                   1,1,1,1,2,2, 1,1,1,2,2,1, 1,1,2,2,1,1, 1,2,2,1,1,1, 2,2,1,1,1,1, 2,1,1,1,1,2, -1,-1,-1,-1,-1,-1};
#if 0
   double pa6b[]={ 1,-1, 1,-1, 1,-1,  /* q31- */
                   0, 0, 0,
                   0, 0, 0, 0, 0, 0,0};
#else
   double pa6b[]={ 0.295411238, 0.295411238, 0.295411238, 0.295411238, 0.295411238, 0.295411238,
                  -0.365148371,-0.365148371,-0.365148371,
                  -0.112837052,-0.112837052,-0.112837052,-0.112837052,-0.112837052,-0.112837052,0};
#endif

/* # x_3 (weight, coeff, |sN...s0>, (symfactor)): E=-2.80277564 q31- m=1/3
   #  0.518233399  0.719884295 |uududd> (12)  4  4  1 idx=1 0.207812696
   #  0.458012574  0.676766262 |ududud> ( 2)  6  0  3 idx=2 0.478546013
   #  0.023754028 -0.154123417 |uuuddd> ( 6)  2  4  3 idx=0 0.062920622
      unpassende Symmetrie!?
   #  0.785410197  0.886233714 |ududud> ( 2)  6  0  3 idx=2 q00+ .62666186888553933938
   #  0.200000000 -0.447213595 |uududd> (12)  4  4  1 idx=1      .12909944472925513742
   #  0.014589803  0.120788258 |uuuddd> ( 6)  2  4  3 idx=0      .04931159983660802569
   */

#define Pa6c1 0.207812696
#define Pa6c2 0.478546013
#define Pa6c3 0.062920622
#define Pa6c4 0.12909944472925513742
#define Pa6c5 0.62666186888553933938
#define Pa6c6 0.04931159983660802569
   int pstate6c[]={1,1,2,1,2,2, 1,2,1,2,2,1, 2,1,2,2,1,1, 1,2,2,1,1,2, 2,2,1,1,2,1, 2,1,1,2,1,2,
                   2,2,1,2,1,1, 2,1,2,1,1,2, 1,2,1,1,2,2, 2,1,1,2,2,1, 1,1,2,2,1,2, 1,2,2,1,2,1,
                   1,2,1,2,1,2,
                   2,1,2,1,2,1,
                   1,1,1,2,2,2, 1,1,2,2,2,1, 1,2,2,2,1,1,
                   2,2,2,1,1,1, 2,2,1,1,1,2, 2,1,1,1,2,2,  -1,-1,-1,-1,-1,-1};
#if 1
   double pa6c[]={ Pa6c1,-Pa6c1, Pa6c1,-Pa6c1, Pa6c1,-Pa6c1, /* q31 state */
                  -Pa6c1, Pa6c1,-Pa6c1, Pa6c1,-Pa6c1, Pa6c1,
                   Pa6c2,-Pa6c2,
                  -Pa6c3, Pa6c3,-Pa6c3, Pa6c3,-Pa6c3, Pa6c3,0};
#else
   double pa6c[]={-Pa6c4,-Pa6c4,-Pa6c4,-Pa6c4,-Pa6c4,-Pa6c4, /* q00 state */
                  -Pa6c4,-Pa6c4,-Pa6c4,-Pa6c4,-Pa6c4,-Pa6c4,
                   Pa6c5, Pa6c5,
                   Pa6c6, Pa6c6, Pa6c6, Pa6c6, Pa6c6, Pa6c6,0};
#endif

   int *pstate=NULL, *ring=NULL;  /* projection state, list of ring sites */
   double *pa=NULL;               /* projection state coefficients */


   if (nn==24) { ring=ring4; vring=1, lring=4, nring=4; }
   if (nn==24) printf(" model state overlap for the squagome model (2 0 0 2)\n");
   if (nn==24 && abs(nu-nd)==16) { pstate=pstate4a; pa=pa4a; }
   if (nn==24 && abs(nu-nd)== 8) { pstate=pstate4b; pa=pa4b; }

   if (nn==27) { ring=ring6_3; vring=3, lring=6, nring=3; }
   if (nn==27) printf(" model state overlap for the kagome model (3 0 0 3)\n");
   if (nn==27 && abs(nu-nd)==21) { pstate=pstate6a; pa=pa6a; } /* m=7/9 EA>1 */
   if (nn==27 && abs(nu-nd)==15) { pstate=pstate6b; pa=pa6b; } /* m=5/9 EA=1 */
   if (nn==27 && abs(nu-nd)== 9) { pstate=pstate6c; pa=pa6c; } /* m=3/9 EA=1 */

   if (nn==36) { ring=ring6_4; vring=3, lring=6, nring=4; }
   if (nn==36) printf(" model state overlap for the kagome model (6 0 2 2)\n");
   if (nn==36 && abs(nu-nd)==20) { pstate=pstate6b; pa=pa6b; } /* m=5/9 EA=1 */
   if (nn==36 && abs(nu-nd)==12) { pstate=pstate6c; pa=pa6c; } /* m=3/9 EA=1 */

   if (nn==45) { ring=ring6_5; vring=3, lring=6, nring=5; }
   if (nn==45) printf(" model state overlap for the kagome model (15 0 4 1)\n");
   if (nn==45 && abs(nu-nd)==25) { pstate=pstate6b; pa=pa6b; } /* m=5/9 EA=1 */

   r1=r3=r4=0; i1=i2=i3=i4=i5=i6=j1=j2=j3=0; i5s=0; n3=0; r2=0;
   if (pstate && pa && ring) /* dont do stupid things */
   for (k2=k=0l;k<n1;k++) {
     b1=getl1(k);             /* get symmetric configuration */
     /* printf("\n |%s> (%ld)",ccfg(b1),n2); */

     b_getbase(b1,&b2); n2=b2.norm2;
     for (k3=0;k3<b2.num;k3++) { /* sum over all symmetric configurations */ 
      ms=b2.is[k3]; b1=b2.ib[k3]; n2=1;

      rr=0; i6=0; /* sum all variants */
      for (j1=0;j1<vring;j1++) {   /* try 1 type of square crystal */
        i4=1;                  /* flag for localized state */
        r5=1;                  /* projection = product of rings */
        for (j2=0;j2<nring && i4;j2++) {
          for (j4=0;pstate[j4*lring]>=0;j4++) { /* ring_projection? */
            for (j3=0;j3<lring;j3++) { /* one localized magnon-ring */
              i1=ring[j1*nring*lring+j2*lring+j3]; /* check site, break if not pstate[j2] */
              if (geti(b1,i1)!=pstate[j3+lring*j4]) break;
            } if (j3>=lring) break; /* nonzero projection found */
          }
          if (pstate[j4*lring]>=0) r5*=pa[j4]; /* projection  */
          else { r5=0; i4=0; break; }
        }
        if (i4) { rr+=r5; i6++; } /* found fitting model state */
      }
      r5=rr /* /sqrt(1.0*i6)??? */; /* mean value q=0 ? */
      rr = M_f2d(vec_e(v0,k)) * ms/sqrt(b2.norm2*1.0);   /* coefficient of config b1 */
      if (i6) {
        /* sort ising energies */
        i5+=i6; i5s+=n2*i6; /* number of projected states */
        r1+= Norm(rr); /* weight of projected states */
        r3+= Norm(r5)*n2/* /i6 */; /* ??? weight of projecting state */
        r2+= r5*rr*sqrt(1.0*n2); /* projection ??? */
      } else { /* maximum non projected configuration */
        if (Norm(rr)>r4) { r4=Norm(rr); k2=k; n3=n2; }
      }
     }
      /* next output the factor of symmetric configurations, Apr03 */ 
   }  
   printf(" num_projected_sym_states %d\n",i5);
   printf(" num_projected_states     %ld\n",i5s);
   printf(" weight_projected_states      %12.8f\n",r1);
   printf(" weight_projecting_states b*b %12.8f ?\n",r3);
   printf(" projection^2         (a*b)^2 %12.8f ?\n",Norm(r2));
   if (r3) /* division by zero */
   printf(" overlap=sqrt((a*b/|b|)^2)    %12.8f\n",sqrt(Norm(r2)/r3));
   printf(" max non_projected state %12.9f |%s> (%ld)",r4,ccfg(getl1(k2)),n3);
 }
#endif
 return 0;
}


/* --------------------------------------------------------------------
 * -------- output coefficients of eigenvector v0 to file *f ----------
 *    on big vectors output the most important ones (big wighted)
 * --------------------------------------------------------------------
 * sort coefficients v0_i, and store largest to r[MAXx-1] 
 * mode=0 maximum values
 * mode=1 minimum values
 * x_p=0  only sorted
 * x_p=1  
 * x_p=3  also output norm*num_same_koeff
 */
int x_out(FILE *f,tvec *v0,int mode){
    ULong i,j,k,ix[MAXx],in[MAXx],ii=0l;cplx r1; /* index???, anzahl */
    double n2; tbase b1;
    static double rq[MAXx];    /* mittelwert, abweichungsquadrat */
    double rr,ep=1.e-5;
    ULong l1=itime();
    static struct sbase b2; b2.o=0;
    if (mpi_i==0 && mpi_n>1) E0("missing mpi code on x_out");
    MPI_Barrier(MPI_COMM_WORLD);
    if (mpi_n) return 0;
    if (!mode) modeloverlap(v0);  /* special types of overlaps */
    if (x_max>MAXx) { x_max=MAXx; ERR(800); }
    if (!x_max) return 0;	/* x_max>0 ! */
    for (i=0l;i<x_max;i++) {ix[i]=n1+1l;in[i]=0l;rq[i]=0;}  /* initial. */
    for (i=0l;i<n1;i++) { 
      if (x_p>0) {			/* collect same coefficients */
        for (j=0l;j<x_max;j++) if (ix[j]<n1)
      	if ((rr=fabs(M_f2d(vec_e(v0,i))-M_f2d(vec_e(v0,ix[j]))))<ep) {
          /* if we dont have norm2[] dont waste time for norm2-calculation */
          in[j]++; if (rq[j]<rr) rq[j]=rr; break;
      	}
      	if (j<x_max) continue; else ii++;
      }	
      for (j=0l;j<x_max;j++) {		/* max (min) coefficients */
         if (ix[j]<n1) {
           if (mode==1) {if ( fabs(M_f2d(vec_e(v0,i)))>ep+fabs(M_f2d(vec_e(v0,ix[j]))) ) break;}
           else         {if ( fabs(M_f2d(vec_e(v0,i)))+ep<fabs(M_f2d(vec_e(v0,ix[j]))) ) break;}
         }
         if (j>0l) {
           ix[j-1]=ix[j];  /* not very well, ToDo: use pointers */
           in[j-1]=in[j];
           rq[j-1]=rq[j];
         }
         ix[j]=i;
         in[j]=1l;
         rq[j]=0.0;
      }
    } 
    r1=M_f2d(vec_e(v0,ix[x_max-1l]));
    if (r1!=0.0) r1=fabs(r1)/r1;  /* signum (phase) */
    if (!mpi_i)
      fprintf(f,"\n x_%lu (weight, factor, coefficient, diff, |sN...s0>, (symfactor)):\n",ii);
    j=0l; rr=0;
    for (k=0l,j=1l,i=x_max-1l;j<=x_max;j++,i--) if (ix[i]<n1) {
       if (x_p>2) {
         fprintf(f," %12.9f",D(Norm(M_f2d(vec_e(v0,ix[i])))*(double)in[i]));/* 12.9f/16.9e */
         fprintf(f," %6lu",in[i]);}
       rr+=Norm(M_f2d(vec_e(v0,ix[i])))*(double)in[i];
       fprintf(f," %s",outc(M_f2d(vec_e(v0,ix[i]))*r1));
       fprintf(f," %5.0e",D(rq[i]));
       b1=getl1(ix[i]);          
       fprintf(f," |%s>",ccfg(b1));
       /* next output the factor of symmetric configurations, Apr03 */ 
       b_smallest(b1,&b2); n2=b2.norm2;
       fprintf(f," (%4g)",n2);

       /* get pair correlations */
       get_dcorr(b1, 2, NULL); /* print out numpairs(uu+dd) = ising ergs */

       fprintf(f," idx=%ld\n",ix[i]);k+=in[i];
    }  fprintf(f," %12.9f %6lu %12s %5.0e <= sum",rr,k,"eps:",ep);
    printf("\n x_out.t= %.2f m", (itime()-l1)/60.);
    MPI_Barrier(MPI_COMM_WORLD);
    return(0);
}


/* ---- sort and count the number of different coefficients --- */
int x_sort(FILE *f,tvec *v0,tvec *v1){
  ULong i,j,h,t1; double r1,r2,r3,r4; cplx v;  int blk;
  if(!x_max) return 0;	/* x_max>0 ! */
  if(A_ptr(v1)==0 || A_ptr(v0)==0
   || v0->v[0]==0 ||  v1->v[0]==0) return(0);
  if (mpi_i==0 && mpi_n>1) E0("missing mpi code on x_sort");
  MPI_Barrier(MPI_COMM_WORLD);
  t1=itime();
  for(blk=0;blk<pt_n;blk++)
  for(i=0l;i<b_len[blk];i++) v1->v[blk][i]=v0->v[blk][i]; /* copy vector */
  /* --- sortieren nach absoluter Laenge --- bug=arrayueberschreitung !!! */
  /* --- shellsort (R.Sedgewick, Algorithmen in C, Addison-Wesley) --- */
  for(h=1;h<=n1/9;h=3*h+1);	/* Distanzen-Folge ...,364,121,40,13,4,1 */
  for(;h>0;h/=3) for(i=h+1;i<n1;i++) {	/* Bug: i<=n1 !!!??? */
    v=M_f2d(vec_e(v1,i)); j=i;
    while(j>h && fabs(M_f2d(vec_e(v1,j-h)))<fabs(v)){
       *vec_pe(v1,j)=vec_e(v1,j-h); j-=h; }
    *vec_pe(v1,j)=M_d2f(v);
  }						/* sorting complete */
  t1=itime()-t1; fprintf(f,"\n x_sort.t= %.2f m", t1/60.);

  for(r2=1e-8;r2>1e-10;r2*=0.25){            /* r2=1e-5,1e-6,..,1e-14 */ 
    r1=fabs(M_f2d(vec_e(v1,0)));r4=1.0;    /* r2=epsilon surroundings */
    for(i=j=1l;i<n1;i++){ 
      r3=fabs(fabs(M_f2d(vec_e(v1,i)))-r1);	/* neighbour distance */
      if(r3>r2){ j++;                 /* sum as same for distance<eps */
                 if(r3<r4)r4=r3; }   /* store smallest distance > eps */
      r1=fabs(M_f2d(vec_e(v1,i))); 
    }
    fprintf(f,"\n A(v,eps=%.0e)=%lu eps2=%.0e",r2,j,r4);
  }
  t1=itime()-t1; fprintf(f,"\n x_sortA.t= %.2f m", t1/60.);
  return(0);
}
/* =================== begin <v0|Op|v1> ============================== */
/* --------------------------------------------------------------------
 * store r*|cfg_i> to hbuf (norm2 too???)
 * this function is used to return the factor and newcfg from Operator(cfg)
 * fast version for diagonal part ???
 * ToDo: improve nhm from this function? sort to different y-blocks
 *  symmetrize and get index later, block wise! (using CPU cache)
 *  - it does not work for Sud and nu==nd if op_XXij isnt u-d-symmetric
 *  - so make your operators u-d-symmetric OR add 2nd loop for kud in wop
 */
// inline 
void puthbuf(thbuf *hbuf, cplx r, tbase bj){
  int ii;  if (r==0) return;
#ifdef CONFIG_ANSATZ
  if(!ansatz_test(bj)) return;
#endif
  /* ToDo: later we should store r into blockno(kj) */
  for(ii=0;ii<hbuf->n;ii++) /* check if (kj) is already stored */
    if(hbuf->el[ii].bj==bj){ hbuf->el[ii].rr+=r;
/*  printf("\n  add term r=%8g +=%8g ii=%d",D(r),D(hbuf->el[ii].rr),ii); */
    return; }
  /* ToDo: sort for faster check if stored??? except diagonal element */
/*  printf("\n  new term r=%8g",D(r)); */
  hbuf->el[hbuf->n].rr=r;                /* store new value */
  hbuf->el[hbuf->n].bj=bj;
  hbuf->n++; if(hbuf->n>=NZXMAX)ERR(9200);
  return;
}

/* this is the replacement for wopij, wopij_paras and wopij_block */
/* to transfer the arguments of the operator to the threads
 * this structure is used, later it could be part of thbuf structure */
struct s_wop_args {  /* not thread save */
  void (*op)(void *,int,int *,thbuf *);  /* operator for n sites */
  void *para;  /* pointer to possible parameters */
  int  num_sites; /* num of sites the operator depend on (example 2) */
  int  sym_tupel; /* number of symmetric tupels */
  int  iw[NN*MaxSym+MaxSym*NN*NN];    /* array of tupels*sites (2D-array) */
} wop_args;

/* store symmetric tupel, expand iw[0..num_sites*num_tupels] */
void store_sym_tupel(){
  int i,j,k;
  wop_args.sym_tupel=1;
  if (!wop_args.num_sites) return; /* operator not depend on sites */
  for (i=0;i<sym.numsym;i++) { /* try all symmetries */
    /* get new tupel */
    for (j=0;j<wop_args.num_sites;j++) /* all sites of one tupel */
      wop_args.iw[wop_args.sym_tupel*wop_args.num_sites+j]=
           sym.i[i][wop_args.iw[j]];
    /* check if tupel already stored */
    for (k=0;k<wop_args.sym_tupel;k++) {
      /* compare stored tupels with new tupel */
      for (j=0;j<wop_args.num_sites;j++) /* all sites of one tupel */
        if (wop_args.iw[wop_args.sym_tupel*wop_args.num_sites+j]
          !=wop_args.iw[                 k*wop_args.num_sites+j]) break;
      if (j==wop_args.num_sites) break; /* tupel already stored */
    }
    if (k==wop_args.sym_tupel) { /* not stored, we have a new! */
      wop_args.sym_tupel++;
    }
  }
#ifdef CONFIG_S1SYM
  if (sym.num_lm!=nn)  /* LM-symmetries (S>=1) */
  { /* try all LM symmetries,
     * better store only pairs with different configs + factor */
    int k=0, loop[NN], nold=wop_args.sym_tupel, ni;
    for (ni=0;ni<nold;ni++) {
      /* get number of new tupels */
      for (j=0;j<wop_args.num_sites;j++){ /* all sites of one tupel */
        loop[j]=wop_args.iw[ni*wop_args.num_sites+j];
        /* reset to LM representator */
        loop[j]=wop_args.iw[ni*wop_args.num_sites+j]=sym.lm[loop[j]];
      }
      /* printf("\n sym tupel %3d: %3d %3d", ni,loop[0],loop[1]); */
      /* generate all symmetric pairs from loop-tupel */
      for (j=0;j<wop_args.num_sites;){ /* start with site 0 as the lowest digit of a number */
        k=sym.ln[loop[j]]; /* increase digit (successor) */
        /* check for roll over, like 9 -> 0, to next higher digit */
        if  (k>=nn) { loop[j]=sym.lm[loop[j]]; j++; continue; }
        loop[j]=k;  /* next digit, like 0 -> 1 -> 2 -> ... -> 9 */
        /* copy from loop-tupel */
        for (k=0;k<wop_args.num_sites;k++)
          wop_args.iw[wop_args.sym_tupel*wop_args.num_sites+k]=loop[k];
        /* printf("\n  LM tupel %3d: %3d %3d", wop_args.sym_tupel,loop[0],loop[1]); */
        if (wop_args.sym_tupel==NN*MaxSym+MaxSym*NN*NN-1)
          E0("store_sym_tupel overflow, dont trust the result");
        if (wop_args.sym_tupel<NN*MaxSym+MaxSym*NN*NN)
          wop_args.sym_tupel++;
        j=0;
      }
    }
  }
#endif
}

/* operator for n-sites void-function for pthreads
 */
void wop_block( void *ptr ){
  ULong i1,i2=0; tbase ib1,ib2; int j,n;
  cplx r;          /* matrix element of OP */
  double rh=0.0;   /* expectation value <v0|OP|v0> */
  /* ULong cnt[5]={0,0,0,0,0}; */
  struct sbase *b2, *b1; thbuf *hbuf=(thbuf *)ptr;
  int block=hbuf->block;
  b1=(struct sbase*)malloc(sizeof(struct sbase)); if(!b1) F0("malloc");
  b2=(struct sbase*)malloc(sizeof(struct sbase)); if(!b2) F0("malloc");
  b1->o=b2->o=0;
  for(i1=0;i1<b_len[block+mpi_i*pt_n];i1++){ /* one block of vector per thread */
    ib1=getl1(i1+b_ofs[block+mpi_i*pt_n]-node_ofs);  /* get config from base index */ 
    b_smallest(ib1,b1); /* need norm2 = orbit length (slow) */ 
    hbuf->n=1;
    hbuf->el[0].rr=0;
    hbuf->el[0].bj=ib1;  /* diagonal element first */
    /* for (kud=0;kud<((nu==nd && sym.wud)?2:1);kud++) */
    {
      for(j=0;j<wop_args.sym_tupel;j++) { /* sum over all symmetric tupel */
        /* now call the user function which calls puthbuf(),
         * ToDo: dynamic-lib dlfcn.h ??? or call set_para_for_op() */
        (*(wop_args.op))(  wop_args.para, wop_args.num_sites,
                         &(wop_args.iw[ j*wop_args.num_sites]), hbuf);
      }
    }
    /* now hbuf should be filled with results from user-operator */
    /* 1st take the diagonal element seperately (fast) */
    r = hbuf->el[0].rr
       * M_f2d(hbuf->v1->v[block][i1]); /* vec_e(v1,i1+b_ofs) */
/*     printf("\n     diagonal r=%g %s terms=%d",
        D(hbuf->el[0][0].rr),ccfg(ib1),hbuf->n[0]); */
    /* now take the off-diagonal elements (slow) */
    for(n=1;n<hbuf->n;n++){
        ib2=hbuf->el[n].bj;
        b_smallest(ib2,b2); /* ToDo: buffer it to have less calls? */
        /* LM-systems, plaquette, ToDo: avoid that */
        if (phase_eq(b2->sign,0)) continue;
        i2=b2i(b2->scfg);  /* get index, segfault if i2>=n1 */
        /* i2>=n1 happens on LM-systems v2.26 (plaquette), do avoid it */
        r+=hbuf->el[n].rr * M_f2d(vec_e(hbuf->v1,i2)) * b2->sign
         *((b1->norm2!=b2->norm2)?sqrt((double)b1->norm2/(double)b2->norm2):1);
        /* if (b1->norm2!=b2->norm2) cnt[3]++; about 1..2% */
    }
    rh += D( Conj(r)*M_f2d(hbuf->v0->v[block][i1])
          /(wop_args.sym_tupel/* *l1.norm2 */) );
    /* was: vec_e(v0,i1+b_ofs) */
  }
  // printf("cnt %lu %lu %lu %lu\n",cnt[0],cnt[1],cnt[2],cnt[3]);
  hbuf->rx= rh; /* give back real(double) result <v0|OP|v0> */
  free(b1); free(b2);
}

/*  ------------------------------------------------------------
 *  use this function to calculate expectationvalues <v0|op|v1>
 *  of your own operator void op(void *param,int i1,int i2,hbuf)
 *  with hbuf=op(cfg,param)
 *  example_call: wop(&op_sisj,&para,num_sites,sites,v0,v1);
 *  =>  (void (*)(void *,int,int *,thbuf *)) &op_sisj
 *  return double (cplx until v2.26, but expectation values are always real)
 */
/* new */
double wop( void (*op)(void *,int,int *,thbuf *),             /* <v1|op(i,j)|v0> */
            void *para, int num_sites, int *sites,
            tvec *v0, tvec *v1){
  thbuf *(hbuf)[B_NUM];            /* one buffer per thread */
  double rh=0.0;                     /* sum over threads */
  int i,block;                     /* thread counter */
#ifdef CONFIG_MPI
#warning "missing mpi code for nondiag terms on wop()"
  // if (mpi_n>1) E0("missing mpi code for nondiag terms on wop()");
#endif
  wop_args.op=op; /* transfer parameters to threads */
  wop_args.para=para;
  wop_args.num_sites=num_sites;
  for (i=0;i<num_sites;i++)
    wop_args.iw[i]=sites[i];  /* copy sites of operator */
  store_sym_tupel(); /* do this before calling wop_block */
  for(block=0;block<pt_n;block++){ if ( b_len[block]==0 ) continue;
    hbuf[block]=alloc_hbuf();
    hbuf[block]->block=block; hbuf[block]->v0=v0;
    hbuf[block]->v1=v1;
    hbuf[block]->rx=0.0;
// #ifdef CONFIG_PTHREAD
    parallel_start( block, wop_block, hbuf[block] );
//#else
//    wop_block( hbuf[block] );
//#endif
  }
  for(block=0;block<pt_n;block++){ if ( b_len[block]==0 ) continue;
    parallel_wait(block);
    rh+= D(hbuf[block]->rx);
    free_hbuf(hbuf[block]); 
  }
#ifdef CONFIG_MPI
  { double add=rh, sum=0;
    MPI_Allreduce(&add, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    rh=sum; }
#endif
  return(rh);
}

/* =================== end <v0|Op|v1> ============================== */
/*
 *  example operator S(i).S(j) = op_sisj for wop calculation
 *  decouples operator from vector and symmetry 
 *  and makes the program more flexible for users to create new operators
 *  call_by wop(&op_sisj,NULL,i1,i2,v0,v1);
 *   this operator is symmetric against spin inversion (u <=> d)
 *
 *  Is the ud-symmetry correct taken into account???
 *  ex: 16q0 3u4d q=(kx/2pi=1/4,ky/2pi=2/4,0-5-2-7=1/4)
 *   wop= SiSj=  0.315944 -0.013741  0.008672 -0.013741 -0.032685  15s/PM600MHz x+z=24s
 *   new= S0Sj=  0.315944 -0.013741  0.008672 -0.013741 -0.032685  13s/PM600MHz
 *   old= S0Sj=  0.037470 -0.007058  0.012980 -0.007058 -0.019535  20s/PM600MHz
 *  tU N=8 see readme.doc
 *  octahedron-J1-J2 (s=5/2)*(N=6) num_intrabonds=2s*2s=25 (since v2.22 Oct04)
 *     Sz=0+ E0=-45.0 SiSj=0.35=s*(s+1)/(num_intrabonds) for intra-spin i-j
 *     SMag=0.2
 */
/* new version */
void op_sisj(void *para,int ns,int *iw,thbuf *hbuf){/* hbuf=OP*hbuf */
  tbase ib1,ib2; int z; double r;
  if (para) ; /* avoid Warning, do not care */
  if (ns) ;   /* avoid Warning, do not care */
  ib1=hbuf->el[0].bj;  /* argument = |cfg> */
  z=zizj(ib1,iw[0],iw[1]); if(z==0) return; r=0;
  if(iw[0]==iw[1]) r+=0.75;else {
    if(z>0)        r+=0.25;  /* ZiZj */
    else {         r-=0.25;  /* SiSj */
      ib2=Flip(ib1,iw[0],iw[1]);
      puthbuf(hbuf,(cplx)0.5,ib2); /* put nondiagonal result */
    }
  }
  puthbuf(hbuf,(cplx)r,ib1); /* put diagonal result */
}
/*
 *  example operator Sxy(i).Sxy(j) = op_xixj for wop calculation
 *  decouples operator from vector and symmetry 
 *  and makes the program more flexible for users to create new operators
 *  call_by wop(&op_xixj,NULL,i1,i2,v0,v1);
 *   this operator is symmetric against spin inversion (u <=> d)
 */
void op_xixj(void *para,int ns,int *iw,thbuf *hbuf){/* hbuf=OP*hbuf */
  tbase ib1,ib2; int z;
  if (para && ns) ; /* avoid Warning, do not care */
  ib1=hbuf->el[0].bj;  /* argument = |cfg> */
  z=zizj(ib1,iw[0],iw[1]);
  if(iw[0]==iw[1] && z!=0) { puthbuf(hbuf,(cplx)0.5,ib1); return; } /* easy case */
  if(z<0) { /* SxiSxj+SyiSyj */
      ib2=Flip(ib1,iw[0],iw[1]);
      puthbuf(hbuf,(cplx)0.5,ib2); /* put nondiagonal result */
  }
}
/*
 *  example operator Sz(i)Sz(j) = op_zizj for wop calculation
 *  decouples operator from vector and symmetry 
 *  and makes the program more flexible for users to create new operators
 *  call_by wop(&op_zizj,NULL,i1,i2,v0,v1);
 *   this operator is symmetric against spin inversion (u <=> d)
 */
/* new version */
void op_zizj(void *para,int ns,int *iw,thbuf *hbuf){/* hbuf=OP*hbuf */
  tbase ib1; int z;
  if (para) ; /* avoid Warning, do not care */
  if (ns) ;   /* avoid Warning, do not care */
  ib1=hbuf->el[0].bj;  /* argument = |cfg> */
  z=zizj(ib1,iw[0],iw[1]); if(z==0) return;
  puthbuf(hbuf,(cplx)z*0.25,ib1); /* put diagonal result */
}

/*
 *  example operator sum_s c(+,i,s)c(-,j,s)+h.c. = op_cicj for wop calculation
 *  decouples operator from vector and symmetry 
 *  and makes the program more flexible for users to create new operators
 *  call_by wop(&op_cicj,NULL,2,sites,v0,v1);
 *   this operator is symmetric against spin inversion (u <=> d)
 *   sample: tU-chain N=8 nud=4,4 t=-1 U=8 e0=... cc(0,1)=... (see m_tU.c)
 */
#if CM(0,1,1)    /* for tJ and tU model only */
void op_cicj(void *para,int ns,int *iw,thbuf *hbuf){/* hbuf=OP*hbuf */
  tbase ib1, ib2; int s, ip1, ip2, ii, is, sign, ia, ib;
  if (para && ns!=2) ; /* avoid compiler warning, do not care */
  ib1=hbuf->el[0].bj;  /* argument = |cfg> */
  if (iw[1]>iw[0]) { ia=iw[0]; ib=iw[1]; } /* order for signum */
  else             { ib=iw[0]; ia=iw[1]; }
  ip1=geti(ib1,ia); /* get site 0, (0: .., 1: .u, 2: d., 3: du) */
  ip2=geti(ib1,ib); /* get site 1 */
  ii=ip1^ip2;
  if (ia==ib && ii>0) { /* simplifies to n(i,s)=c(+,i,s)c(-,i,s) */
    puthbuf(hbuf,(cplx)((ii<3)?1:2),ib1); /* put diagonal result */
    return;
  }
  if ( ! (CM( ii , ii>0 && ii<3 , ii )) ) return;
  is=parity(ib1 & p_mask[ia][ib]); /* fermionic signum without sites */
  for (s=0;s<2;s++) if (1&(ii>>s)) { /* 0=up, 1=down */
    sign = (( (s==0) ? ip1&2 : ip2&1 ) ? -is : is ); /* sign correction */
    ib2=tbase_xor(tbase_xor(ib1,ia,s),ib,s);
    puthbuf(hbuf,(cplx)sign,ib2); /* put nondiagonal result */
  }
}
#endif
/*
 *  example operator  B(i,j)*B(k,l)
 *      B(i,j)=sum_s c(+,i,s)c(-,j,s)+h.c. = op_cicj for wop calculation
 *  decouples operator from vector and symmetry 
 *  and makes the program more flexible for users to create new operators
 *  call_by wop(&op_cicjckcl,NULL,4,sites,v0,v1);
 *   this operator is symmetric against spin inversion (u <=> d)
 *   sample: tU-chain N=8 nud=4,4 t=-1 U=8 e0=... cc(0,1)=... (see m_tU.c)
 */
#if CM(0,1,1)    /* for tJ and tU model only */
void op_cicjckcl(void *para,int ns,int *iw,thbuf *hbuf){/* hbuf=OP*hbuf */
  /* for simplification we use op_cicj to build the product */
  /* ok, code is not looking fine, so it subject of change */
  thbuf hb1, hb2; int site[2], i, j; cplx r;   /* used to build the product */
  /* thbuf *htmp=alloc_hbuf(); free_hbuf(htmp); */
  if (para && ns) ; /* avoid compiler warning, do not care */
  hb1.el[0].bj=hbuf->el[0].bj;  /* copy hbuf to hb1 */
  hb1.el[0].rr=hbuf->el[0].rr;  /* copy hbuf to hb1 */
  hb1.n=1;
  site[0]=iw[0];
  site[1]=iw[1];
  op_cicj(NULL,2,site,&hb1);  /* calling puthbuf(), result=hb1 */
  site[0]=iw[2];
  site[1]=iw[3];
  for (i=0; i<hb1.n; i++)  /* for all terms in hb1 */
  {
    hb2.el[0].bj=hb1.el[i].bj;  /* copy hb1[i] to hb2 */
    hb2.el[0].rr=0;
              r =hb1.el[i].rr;  /* copy prefactor */
    hb2.n=1;
    op_cicj(NULL,2,site,&hb2); /* calling puthbuf(), result=hb2 */
    for (j=0; j<hb2.n; j++) {  /* store product hb1*hb2 to hbuf */
      puthbuf(hbuf, (cplx)r*hb2.el[j].rr, hb2.el[j].bj);
    }
  }
}
#endif

/*
 *  example operator  P(i,i+1)+P^-1(i,i+1): (i,j,k,l) => (j,k,l,i)+(l,i,j,k)
 *      Rotation of sites
 *  decouples operator from vector and symmetry 
 *  and makes the program more flexible for users to create new operators
 *  call_by wop(&op_rot4,NULL,4,sites,v0,v1);
 *   this operator is symmetric against spin inversion (u <=> d)
 *   sample:
 * ToDo: how can this be used in hamilton_nhv?
 */
#if CM(1,0,0)    /* for J (for tU model sign must be calculated) */
void op_rot4(void *para,int ns,int *iw,thbuf *hbuf){/* hbuf=OP*hbuf */
  tbase ib1, ib2; int ip1, i;
  if (para) ; /* avoid compiler warning, do not care */
  ib2 = ib1 = hbuf->el[0].bj;  /* copy the argument = |cfg> */
  for (i=0; i<ns; i++) {
    ip1 = geti(hbuf->el[0].bj, iw[i]);      /* take site i */
    ib1 = puti(ib1, iw[(i   +1)%ns], ip1);  /* put to site i+1 */
    ib2 = puti(ib2, iw[(i+ns-1)%ns], ip1);  /* put to site i-1 */
  }
  puthbuf( hbuf, (cplx)1.0, ib1); /* rotation */
  puthbuf( hbuf, (cplx)1.0, ib2); /* inverse rotation */
}
#endif

/* ------------------------------------------------------------------ */
/* ToDo: call hamilton(v0,v1=NULL) (parallel)??? v1=H*v0+0*v1 */
double wh(tvec *v0){			/* expectation value <H> */
  double rh=0.0; ULong i; double rd=0; int blk;
  thbuf *hbuf;
  hbuf=alloc_hbuf();
  if (mpi_n>1) E0("missing mpi code on wh");
  hbuf->v0 = v0;
  for (blk=0; blk<pt_n; blk++){
    hbuf->block = blk;
    /* ToDo: better call H*vblock (new nhv) operator pt_n times? */
    for (i=0; i<b_len[blk]; i++) { 
      // ToDo: add num_ahead + check correctness
      hamilton_nhv(b_ofs[blk]+i,1,hbuf,0);
      rh += D(Conj(M_f2d(v0->v[blk][i]))*hbuf->el[0].rr);
      rd +=   Norm(M_f2d(v0->v[blk][i]))*hbuf->nhn; // wrong el[1].rr ??
    }
  }
  printf("\n traceH=%12.8f",rd);  /* traceH is different from sum_Ising!? */
  free_hbuf(hbuf);
  if (vvv&1) printf(" %s",sclock(~0l));
  return(rh);
}

/* ToDo: multitask, via hamilton() H*v0+v1 -> H*v0+a*v1 (where v1=v0?) */
double whq(tvec *v0,cplx rh){  /* deviation of <H> (H*v-h*v)/h */
     /* ToDo: better call H*vblock operator pt_n times? */
    double rq=0.0; ULong i; int blk;
    thbuf *hbuf;
    if (mpi_n>1) E0("missing mpi code on whq");
    hbuf=alloc_hbuf();
    hbuf->v0 = v0;
    for (blk=0; blk<pt_n; blk++){
      hbuf->block = blk;
      for(i=0; i<b_len[blk]; i++){ 
        hamilton_nhv(b_ofs[blk]+i,1,hbuf,0); // ToDo: check correctness
        rq+=Norm(hbuf->el[0].rr
            -rh*M_f2d(v0->v[blk][i])); 
      }
    }
    rq=sqrt(rq/(double)n1);
    free_hbuf(hbuf);
    if (vvv&1) printf(" %s",sclock(~0l));
    return(rq);
} 

/* -------------------------------------------------------------------
 * mean value <v1|T*v0>/<v0|v0> (v1=v0) and deviation (T|v0>-<T>v0)^2
 * for symmetry T
 *   |k> = q sum_R^l exp(i*k*R) T_R |v>   ( Tx_R |x> = |x+R> ) 
 *       = q sum_m^l exp(i*k*m) T^m |v>   - Fouriertransformation k=2*Pi/l*k
 *   rk  = <v|k>			  q=1/sqrt(sum_m^l exp(i*k*m)^2)=2/l
 *  --- obsolete - removed in v2.27 ---
 */ 

/*
 *  operator S^2 = op_ss for wop calculation (slow!)
 *  decouples operator from vector and symmetry 
 *  and makes the program more flexible for users to create new operators
 *  call_by wop(&op_sxsx,NULL,0,NULL,v0,v1); # i1 and i2 is ignored
 *  terrible slow! why?
 *   this operator is symmetric against spin inversion (u <=> d)
 */
void op_sxsx(void *para,int ns,int *iw,thbuf *hbuf){/* hbuf=OP*hbuf */
  tbase ib1, ib2; int z, j1, j2;
  if (para && ns && iw) ; /* avoid Warning, do not care */
  ib1=hbuf->el[0].bj;  /* argument = |cfg> */
  CM( z=nn, z=nu+nd, {for(z=j1=0;j1<nn;j1++) if(zi(ib1,j1)) z++;});
  puthbuf(hbuf,0.5*z,ib1);          /* Sum_i(S(+,i)S(-,i)) */
  for (j1=0;j1<nn-1;j1++) for (j2=j1+1;j2<nn;j2++) { /* i<j */
    z=zizj(ib1,j1,j2); if(z>=0) continue;
    ib2=Flip(ib1,j1,j2);
    puthbuf(hbuf,(cplx)1,ib2); /* put nondiagonal result */
  }
}

/* ??? faster version */
zahl wss(tvec *v0,tvec *v1){ /* SS=1/2*(SpSm+SmSp)+S0^2 ss=s(s+1) */
    /* newversion: SS=SZ^2+1/2Sum_(SipSjm+SimSjp) if [SZ,H]=0 */
    cplx rq,
      rh=wop(&op_sxsx,NULL,0,NULL,v0,v1)+0.25*(nu-nd)*(nu-nd); /* +Z^2 */
    printf("\n SS= %8.5f",D(rh)); 
    rh=fabs(-1.0+sqrt(1.0+4.0*rh));
    rq=fabs(rh-(int)(fabs(rh)+0.5)); /* allways rh>0 */
    printf(" 2S= %5.3f % 5.0e",D(rh),D(rq)); 
    if(vvv&1)printf(" %s",sclock(~0l));
    return(D(rh));
}
/*
 *  operator S^2 = op_ss for wop calculation
 *  decouples operator from vector and symmetry 
 *  and makes the program more flexible for users to create new operators
 *  call_by wop(&op_sxsx,NULL,0,NULL,v0,v1); # i1 and i2 is ignored
 *  only for bipartit lattices!
 *   this operator is symmetric against spin inversion (u <=> d) ???
 */
void op_jxjx(void *para,int ns,int *iw,thbuf *hbuf){/* hbuf=OP*hbuf */
 // static int sl[8]={ 1,-1, 1,-1, 1,-1, 1,-1}; /* SL defined by hand */
  tbase ib1,ib2; int z,ip1,ip2,j1,j2;
  if (para && ns && iw) ; /* avoid Warning, do not care */
  ib1=hbuf->el[0].bj;  /* argument = |cfg> */
  CM( z=0, z=0, {for(z=j1=0;j1<nn;j1++) if(!zi(ib1,j1)) z++;}); /* sum z=0 */
  puthbuf(hbuf,0.5*z,ib1);          /* Sum_i(S(+,i)S(-,i)) */
  for (j1=   0;j1<nn-1;j1++)
  for (j2=j1+1;j2<nn  ;j2++) { /* i<j */
    ip1=geti(ib1,j1); ip2=geti(ib1,j2);
    if((ip1==0 && ip2==3) || (ip1==3 && ip2==0)) { /* .3 <=> 3. */
      ib2=Flip(ib1,j1,j2); /* (.3) and (3.) */
      z=1-2*(1&(j1+j2)); /* +1 for sublattice A, -1 for sublattice B */
      // z=sl[j1]*sl[j2];    /* +1 for sublattice A, -1 for sublattice B */
      puthbuf(hbuf,(cplx)z,ib2); /* put nondiagonal result */
    }
  }
}

/*  kind of dimer operator, example for chains: <S(i)S(i+2)-S(i-2)S(i)>
 *  operator S(i).S(j)-S(k).S(l) = op_sisjsksl for wop calculation
 *  decouples operator from vector and symmetry 
 *  and makes the program more flexible for users to create new operators
 *  call_by wop(&op_sisjsksl,NULL,4,sites,v0,v1);
 *   this operator is symmetric against spin inversion (u <=> d)
 *  <SiSj-SkSl> is different from <SiSj>-<SkSl> ???, no (just a sample)
 *  example: <ud..du|Op|du..ud>
 *  correct?
 *   m_tU 10 >daten.def, param=10,1,0,1,0.5,0.3333333,0.25,0.1 nud=2,2
 *   E0= 0.49214577 (ok), <s(i)s(i+2)-s(i-2)s(i)> i= 2 -0.445099 (???) 
 *        sum_k= -1 0 -1000 : -0.447823
 *        sum_k= -1 -1000   :  0.000000 ToDo: check for error
 */
void op_sisjsksl(void *para,int n,int *iw,thbuf *hbuf){/* hbuf=OP*hbuf */
   /* in prinzip code could be replaced by two calls to op_sisj */
  tbase ib1,ib2; int j1,j2,z; double r;
  if (para) ; /* avoid Warning, do not care */
  if (n) ;    /* avoid Warning, do not care */
  r=0;
  ib1=hbuf->el[0].bj;  /* argument = |cfg> */
  j1=iw[0];
  j2=iw[1];
  z=zizj(ib1,j1,j2);
  if (z!=0) {
    if(j1==j2) r+=0.75;else {
      if(z>0)  r+=0.25;  /* ZiZj */
      else {   r-=0.25;  /* SiSj */
        ib2=Flip(ib1,j1,j2);
        puthbuf(hbuf,(cplx)0.5,ib2); /* put nondiagonal result */
      }
    }
  }
  j1=iw[2];
  j2=iw[3];
  z=zizj(ib1,j1,j2);
  if (z!=0) {
    if(j1==j2) r-=0.75;else {
      if(z>0)  r-=0.25;  /* ZiZj */
      else {   r+=0.25;  /* SiSj */
        ib2=Flip(ib1,j1,j2);
        puthbuf(hbuf,(cplx)-0.5,ib2); /* put nondiagonal result */
      }
    }
  }
  puthbuf(hbuf,(cplx)r,ib1); /* put diagonal result */
}


/* ToDo: check tU chain N=16 result and speed old/new sym/nosym, remove old
 *    opJ(+,i)=eps(i)*c(-,i,up)*c(-,i,down)    # eps(i)=(-1)^i or sublattice-AB????
 *    opJ(-,i)=eps(i)*c(+,i,down)*c(+,i,up)
 *    opJ(z,i)=1/2*(1-n(i,up)-n(i,down))
 *    [H,opJ(z)]=0, [H,opJ^2]=0  for bipartit lattices
 *   opJ(-,i)opJ(+,i)=c(+,i,down)*c(+,i,up)*c(-,i,up)*c(-,i,down)=niu*nid
 */ 
cplx wjj(tvec *v0,tvec *v1){ /* pseudo spin JJ=J(J+1) */
#if CM(1,0,0)
    if (v0==v1 && v0==0) return 0; /* avoid unused variable warnings */
    return 0;                      /* not defined for Heisenberg model */
#else
    cplx rh=0.0, rq;
    /* Feb04-test tU-N=8-chain 2J=0.001 failed!? see doc/readme.doc */
    rh=wop(&op_jxjx,NULL,0,NULL,v0,v1)+0.25*(nu+nd-nn)*(nu+nd-nn); /* +j0*j0*/
    if (!mpi_i) printf("\n JJ= %8.5f",D(rh));
    rh=-1.0+sqrt(1.0+4.0*rh);rq=fabs(rh-(int)(fabs(rh)+0.5));
    if (!mpi_i) {
      printf(" 2J= %5.3f %6.0e",D(rh),D(rq));
      if (vvv&1) printf(" %s",sclock(~0l));
    }
    return(rh);
#endif
}

/*  -- Warning! Meaning of the operator has changed since v2.21 -- 
 *  example operator (nu+nd)(i)(nu+nd)(j) = op_ninj for wop calculation
 *  decouples operator from vector and symmetry 
 *  and makes the program more flexible for users to create new operators
 *  iw[0]=site1;iw[1]=site2; wop(&op_ninj,NULL,2,iw,v0,v1);
 *   this operator is not symmetric against spin inversion (u <=> d) ???
 */
void op_ninj(void *para,int ns,int *iw,thbuf *hbuf){/* hbuf=OP*hbuf */
  tbase ib1;  int num_n;
  static int tab_n[4]={0,1,1,2};  /* num of spins per site from geti() */
  if (para &&  ns) ; /* avoid Warning, do not care */
  ib1=hbuf->el[0].bj;  /* argument = |cfg> ToDo: gethbuf? */
  num_n=tab_n[geti(ib1,iw[0])]*tab_n[geti(ib1,iw[1])];
  if(num_n)
    puthbuf(hbuf,(cplx)(1.0*num_n),ib1); /* put diagonal result */
}

/*
 *  example operator n(i,si)n(j,sj) = op_nisnjs for wop calculation
 *  decouples operator from vector and symmetry 
 *  and makes the program more flexible for users to create new operators
 *  call_by wop(&op_nisnjs,(void *)(int a[4]={1,1,1,1}),2,iw,v0,v1);
 *  call_by wop(&op_nisnjs,(void *)"1111",              2,iw,v0,v1);
 *   n(i,up)n(j,up): {1,1,1,1} or MACRO WopParam_UpUp (?)
 *   this operator is not symmetric against spin inversion (u <=> d) ???
 *   its only useful for tJ and tU model
 */
void op_nisnjs(void *para,int ns,int *iw,thbuf *hbuf){/* hbuf=OP*hbuf */
  tbase ib1; char *spara=(char *)para; /* spin-Mask up-up or hole-hole etc. */
  ib1=hbuf->el[0].bj;  /* argument = |cfg> ToDo: gethbuf? */
  /*  if (ns<0) { printf(" %d %d %d %d",spara[0],spara[1],spara[2],spara[3]);
      exit(1); } */
  if((geti(ib1,iw[0])&(spara[0]&3))==(spara[1]&3)
  && (geti(ib1,iw[1])&(spara[2]&3))==(spara[3]&3))
    puthbuf(hbuf,(cplx)1.0,ib1); /* put diagonal result */
}

/*  --- occupation operator ---
 *  example operator n(i,s) = op_nis for wop calculation
 *  decouples operator from vector and symmetry 
 *  and makes the program more flexible for users to create new operators
 *  call_by wop(&op_nis,(void *)&s1,1,&site,v0,v1);
 *   this operator is not symmetric against spin inversion (u <=> d) ???
 *   s1 is "1" for n(i,up), "2" for n(i,down) and "3" for nu*nd
 */
void op_nis(void *para,int ns,int *iw,thbuf *hbuf){/* hbuf=OP*hbuf */
  tbase ib1; char *s1=(char *)para;
  if (para && ns) ; /* avoid Warning, do not care */
  ib1=hbuf->el[0].bj;  /* argument = |cfg> */
  if(((geti(ib1,iw[0])&(s1[0]&3))==(s1[0]&3)))
    puthbuf(hbuf,(cplx)1,ib1); /* put diagonal result */
}

/* ------------- Marshall-Peierls phase rule --------------------------------
 * see J.~Richter {\it et. al}, Euro.~Phys.~Lett. {\bf 25}, 545 (1994)
 *     K.~Retzlaff {\it et. al}, Z. Phys. B {\bf 93}, 21-31 (1993)
 * input:  v0=\sum_i a_i |isingstate_i>
 * output:
 *  - number and weight of coefficients violating the marshall sign rule
 *    (states which have the wrong sign)
 *    i.e.  sign(a_i) != ms,   where  ms=marshall_sign(isingstate_i,SL)
 *  mw=<v0|((sign(a_i)==ms)?1:0)|v0>
 *
 * definition of the Marshall sign:
 *  ms=(-1)^(Sa-Ma), where Sa=\sum_{i\in A} S_i, Ma=\sum_{i\in A}
 *                   and A is a sublattice defined in sl[]
 * sign = ((number of down spins in one sublattice) mod 2)*2-1
 * idea: IsingGS, number nn-excitations change sign of other IsingStates
 *    highest IsingExcitation (has lowest koefficient) changes sign first?
 *    highest IsingExcitation grows with N, mrule earlier broken?
 *
 * ToDo:
 * - replace slow b_getbase?
 * - proof for complex vectors !???
 * - SL-definition is not general??? how to make better definition?
 *   I0 = lowest Isingenergy (for bond i, which is J1 or J2, ...)
 *   I1 = energy of 1st Isingexcitation -> set in ini_norm2 (integer)
 *   generalmrule=(-1)^((Isingerg-I0)/(I1-I0))*sign(I0state)
 *   Isingenergy is Bond dependent (Ierg(J1) or Ierg(J2) -> MRule(Ji))
 *   no sublattice definition necessary
 * - does this definition holt for tJ- and tU-model ??? odd N? 
 * - what about SL-mrule (see J1-J2-model)
 * - SL-rule for triangle lattice (generalization?)
 * - learn/get SL from the heaviest state?
 * - mrule on complex vectors senceful?
 *
 * ex: N=8-J1J2-chain (noSBase) J2krit very sensitiv to numeric errors
 *  count small values (<1e-16) extra?
 *  ud=4+4 J2=.0    MR=1.000 70+0+0 0.000 E=-3.65109341
 *  ud=4+4 J2=.249  MR=1.000 70+0+0 0.000 E=-3.28214917 # states=0..0 err<1e-16
 *  ud=4+4 J2=.250  MR=1.000 69+0+1 2e-34 E=-3.28077641 # states=0..8 err<1e-16 dbl v2.27 k=0 or k=-1
 *  ud=4+4 J2=.251  MR=1.000 62+0+8 5e-09 E=-3.27940480 # states=8..8 
 *  ud=3+5 J2=.0    MR=1.000 56+0+0 0.000 E=-3.12841906
 *  ud=3+5 J2=.261  MR=1.000 48+0+8 1e-08 E=-2.80266111
 *  ud=2+6 J2=.0    MR=1.000 28+0+0 0.000 E=-1.80193774
 *  ud=2+6 J2=.334  MR=1.000 20+0+8 2e-07 E=-1.33266748
 *  ud=1+7 J2=.293  MR=0.500  4+0+4 0.500 E= 0.58589322
 *  square 36+4 J2=0.5 MR=0.998989    67470+23920    0.00101114 E=14.17156790
 *  square 34+6 J2=0.5 MR=0.998020  2426500+1411880  0.00198041 E= 6.80165104 5s PM600MHz
 *  square 33+7 J2=0   MR=0.500000  9321780+9321780  0.500000   E=-2.75696607 26s PM600MHz
 *  square 32+8 J2=0.5 MR=0.994331 42885265+34019420 0.00566940 E= 0.10205747 108s PM600MHz 
 *
 * ToDo:
 *  Speed?: operation (IOPS/FLOPS,M/L1/L2-read) needed and time
 */
struct s_mrule_vars { /* data transfer to/from threads */
  int block;
  int *sl;            /* sublattice definition, global */
  ULong j1,j2,j0;
  double r1,r2;
  tvec *v0;
};

/* see *_old, try on 6site sample */
/* void-function for pthreads, much faster v2.27 */
void mrule_block( void *ptr ){
  ULong i,j0,j1,j2; double r1,r2,r3; cplx ai;
  int i1,ms,mu,md,site,block,*sl; 
  struct sbase l2;
  tbase b1;
  struct s_mrule_vars *tbuf=(struct s_mrule_vars *)ptr;
  j0=j1=j2=0; r1=r2=0.0; sl=tbuf->sl; block=tbuf->block;
  for(i=0l;i<b_len[block];i++){
    b1=getl1(i+b_ofs[block]); ms=mu=md=1; l2.norm2=1;
    ai=M_f2d(tbuf->v0->v[block][i]); /* coefficient b_ofs+i of vector v0 */

    /* if symmetry used, MW depend from norm2 */
    l2.num=0; l2.o=0; b_smallest(b1,&l2); /* slow, bad! ToDo: check why! */

    r3=Norm(ai);
    /* calculate the Marshall sign */
    ms=mu=md=1;
    for (i1=0;i1<nn;i1++) if (sl[i1]==0) { /* check sublattice 0 = A */
      site=geti(b1,i1);  /* 0=empty, 1=up, 2=down, 3=up+down */
      if (site==2) md=-md; else
      if (site==1) mu=-mu;
    } ms=md;
    /* the same within symmetric configs? */
    if (D(ms*ai)<0.0){ j1+=l2.norm2; r1+=r3; } else
    if (D(ms*ai)>0.0){ j2+=l2.norm2; r2+=r3; } else
                       j0+=l2.norm2; /* count zero coefficients */
  }   /* loop over all vector elements of the block */
  tbuf->j0=j0;
  tbuf->j1=j1;
  tbuf->j2=j2;
  tbuf->r1=r1;
  tbuf->r2=r2;
  return;
}

int mrule(tvec *v0, int *sl, int num_sl){
  struct s_mrule_vars tbuf[B_NUM];            /* one buffer per thread */
  ULong j1,j2,j0,j3;
  long t1,t2;            /* timer, speed measurement */
  int i1,block; 
  double r1,r2,r3;       /* j3,r3 temporary vars */
  r1=r2=0.0; j0=j1=j2=0;
  if (!mpi_i) printf("\n MRule=");
  t1=itime();
  if(num_sl>1) {
    for(block=0;block<pt_n;block++){ if ( b_len[block]==0 ) continue;
      tbuf[block].block=block;
      tbuf[block].v0=v0;
      tbuf[block].sl=sl;
#ifdef CONFIG_PTHREAD
      parallel_start( block, mrule_block, &(tbuf[block]) );
#else
      mrule_block( &(tbuf[block]) );
#endif
    }
    for(block=0;block<pt_n;block++){ if ( b_len[block]==0 ) continue;
#ifdef CONFIG_PTHREAD
      parallel_wait(block); 
#endif
      j0+=tbuf[block].j0; /* summerize parallel results */
      j1+=tbuf[block].j1;
      j2+=tbuf[block].j2;
      r1+=tbuf[block].r1;
      r2+=tbuf[block].r2;
    }
  }
#ifdef CONFIG_MPI
  { double add[5], sum[5];
    add[0]=j0; add[1]=j1; add[2]=j2; add[3]=r1; add[4]=r2;
    MPI_Allreduce(&add, &sum, 5, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    j0=(ULong)add[0]; j1=(ULong)add[1]; j2=(ULong)add[2]; r1=add[3]; r2=add[4];
     }
#endif
  /* sort it before output for better postprocessing, j3 used as flag */
  j3=0; if (r1>r2) {j3=j1; j1=j2; j2=j3; j3=1; r3=r1; r1=r2; r2=r3;}
  if (!mpi_i) {
    printf(" %7.5e %8ld %ld %8ld %7.5e ex=%csgn(%+.0e) err= % .0e ",
             r2, j2, j0, j1, r1, ((j3)?'+':'-'), D(v0->v[0][0]), 1-r1-r2);
    /* ToDo: for correct sign we need the Ising-Groundstate configuration */
    printf("\n SL: "); for(i1=0;i1<nn;i1++) printf("%c",'A'+sl[i1]);
    printf(" (may set via SL= ABABAB... in input file)");
    t2=itime();
    printf("\n mrule.t=%s",sclock(t2-t1));
    printf(" t+= %s\n",sclock(~0l));
  }
  return 0;
}


/*
 *   operator Sz(i) = op_zi for wop calculation
 *  decouples operator from vector and symmetry 
 *  and makes the program more flexible for users to create new operators
 *  call_by wop(&op_zi,NULL,1,&site,v0,v1);
 *   this operator is symmetric against spin inversion (u <=> d) ???
 */
void op_zi(void *para,int ns,int *iw,thbuf *hbuf){/* hbuf=OP*hbuf */
  tbase ib1; int z;
  if (para && ns) ; /* avoid Warning, do not care */
  ib1=hbuf->el[0].bj;  /* argument = |cfg> */
  z=zi(ib1,iw[0]); if(z==0) return;
  puthbuf(hbuf,(cplx)z*0.5,ib1); /* put diagonal result */
}


/* evl. define zi()-fct for spin1 in modeldef.h! (more general) */
/* only for dimer model, spin 1 chains , spin-1 = 2 Spin-1/2 = (4n+2)+(4n+3)
   O(z,pi,i,j)=<v1|Szi*exp(sum(i*Pi*Szk,k=i+1..j))*Szj|v0> |Si|=1*/
int stringorder(FILE *f1,tvec *v0){  
  ULong i;int j,z,k,ex,iw1,iw2;double r,rh;
  static struct sbase l2;
  if (mpi_n>1) E0("missing mpi code on stringorder");
  l2.o=0;
  for(iw2=1;iw2<nn/4;iw2++){ /* max. abstand */
    iw1=0;rh=0.0;
    for(i=0l;i<n1;i++){
      b_getbase(getl1(i),&l2);r=0;
      if (err_fatal==520) return 0;
      for(j=0;j<l2.num;j++)if(l2.is[j]!=0){
        ex=0;for(k=iw1+1;k<=iw2;k++)ex+=zi(l2.ib[j],4*k+2)+zi(l2.ib[j],4*k+3);
        z=(zi(l2.ib[j],4*iw1+2)+zi(l2.ib[j],4*iw1+3))
         *(zi(l2.ib[j],4*iw2+2)+zi(l2.ib[j],4*iw2+3))
         *(1-(2&ex));
        if(z!=0)r+=(z*Norm(l2.is[j]));}  /* l2... noetig? */ 
      rh+=(.25*r)*Norm(M_f2d(vec_e(v0,i)))/(double)l2.norm2; }
    /* mit kommentar, kann mit shelltools beseitigt werden (awk) */
    fprintf(f1," O(z,pi,%d,%d)= %9.6f",iw1,iw2,rh);
  } printf(" <- string-order\n");
  return 0;
}

/* only for spin 1 chains , spin-1 = 2 Spin-1/2 = (4n+2)+(4n+3)
   O(z,pi,i,j)=<v1|Szi*exp(sum(i*Pi*Szk,k=i+1..j))*Szj|v0> |Si|=1*/
int stringorder2(FILE *f1,tvec *v0){  
  ULong i;int j,z,k,ex,iw1,iw2;double r,rh;
  static struct sbase l2;
  if (mpi_n>1) E0("missing mpi code on stringorder2");
  l2.o=0;
  for(iw2=1;iw2<nn/2;iw2++){ /* max. abstand */
    iw1=0;rh=0.0;
    for(i=0l;i<n1;i++){
      b_getbase(getl1(i),&l2);r=0;
      if (err_fatal==520) return 0;
      for(j=0;j<l2.num;j++)if(l2.is[j]!=0){
        ex=0;for(k=iw1+1;k<=iw2;k++)ex+=zi(l2.ib[j],k)+zi(l2.ib[j],k+nn/2);
        z=(zi(l2.ib[j],iw1)+zi(l2.ib[j],iw1+nn/2))
         *(zi(l2.ib[j],iw2)+zi(l2.ib[j],iw2+nn/2))
         *(1-(2&ex));
        if(z!=0)r+=(z*Norm(l2.is[j]));}  /* l2... noetig? */ 
      rh+=(0.25*r)*Norm(M_f2d(vec_e(v0,i)))/(double)l2.norm2; } 
    /* mit kommentar, kann mit shelltools beseitigt werden (awk) */
    fprintf(f1," O(z,pi,%d,%d)= %9.6f",iw1,iw2,rh);
  } printf(" <- string-order\n");return 0;}

#if defined(N1) && defined(N2) && N1>1  /* old non-sym version */
cplx wdx(tvec *v0){
    ULong i;tbase ib1;int i1,i2,z0;cplx rdx=0.0,r;
    if (mpi_n>1) E0("missing mpi code on wdx");
    for(i=0l;i<n1;i++){ ib1=getl1(i); 
      /* Dimer Order Dx */
      r=0.0;for(i1=0;i1<nn;i1++){
       i2=site_move(i1,1,0,0);z0=zizj(ib1,i1,i2);
       r+=(1-2*((i1+i1/(nn/N1))&1))  /* sublattice A=+1 B=-1 */
         *(M_f2d(vec_e(v0,i))*z0+((z0==-1)?2:0)
          *M_f2d(vec_e(v0,b2i(Flip(ib1,i1,i2)))));} 
      rdx+=2*Norm(0.25*r/nn); } /* (c)?a:b;  a,b != cplx gnu 2.5.8 */
    return rdx;}
    
cplx wdy(tvec *v0){
    ULong i;tbase ib1;int i1,i2,z0;cplx rdy=0.0,r;
    if (mpi_n>1) E0("missing mpi code on wdy");
    for(i=0l;i<n1;i++){ ib1=getl1(i);
      /* Dimer Order Dy */
      r=0.0;for(i1=0;i1<nn;i1++){
       i2=site_move(i1,0,1,0);z0=zizj(ib1,i1,i2);
       r+=(1-2*((i1/(nn/N1))&1))*(M_f2d(vec_e(v0,i))*z0
        +((z0==-1)?2:0)*M_f2d(vec_e(v0,b2i(Flip(ib1,i1,i2)))));} 
      rdy+=2*Norm(0.25*r/nn);}
    return rdy;}

cplx wc(tvec *v0){
    ULong i;tbase ib1;int i1,i2,z0;cplx rc=0.0,r;
    if (mpi_n>1) E0("missing mpi code on wc");
    for(i=0l;i<n1;i++){ ib1=getl1(i);
      /* Collinear Order C (for square j1/j2) 
         = (1/4)*sum_i S(i)S(i+x)+S(i)S(i-x)-S(i)S(i+y)-S(i)S(i-y)
         = (1/2)*sum_i S(i)S(i+x)-S(i)S(i+y)  ???
        */
      r=0.0;for(i1=0;i1<nn;i1++){
       i2=site_move(i1, 1,0,0);z0=zizj(ib1,i1,i2);
       r+=(M_f2d(vec_e(v0,i))*z0+((z0==-1)?2:0)*M_f2d(vec_e(v0,b2i(Flip(ib1,i1,i2)))));
       i2=site_move(i1,-1,0,0);z0=zizj(ib1,i1,i2);
       r+=(M_f2d(vec_e(v0,i))*z0+((z0==-1)?2:0)*M_f2d(vec_e(v0,b2i(Flip(ib1,i1,i2)))));
       i2=site_move(i1,0, 1,0);z0=zizj(ib1,i1,i2);
       r-=(M_f2d(vec_e(v0,i))*z0+((z0==-1)?2:0)*M_f2d(vec_e(v0,b2i(Flip(ib1,i1,i2)))));
       i2=site_move(i1,0,-1,0);z0=zizj(ib1,i1,i2);
       r-=(M_f2d(vec_e(v0,i))*z0+((z0==-1)?2:0)*M_f2d(vec_e(v0,b2i(Flip(ib1,i1,i2)))));
      } rc+=Norm(0.125*r/nn);}
    return rc;}

cplx wz(tvec *v0){
    ULong i;tbase ib1;int i1,z0,sx,zx,sy,zy,sxy,zxy;cplx rz=0.0,r;
    if (mpi_n>1) E0("missing mpi code on wz");
    for(i=0l;i<n1;i++){ ib1=getl1(i);
      /* better ?? ######### <cfg|Op(i,j,k)*v0> ########################### */
      /* Zi=<[1/(2N)Sum(Z(i,i+x,i+x+y)-Z(i,i+x+y,i+y))]^2>, Zijk=8(ez*Gijk)
       * Eijk=Si(Sj x Sk), Gijk=(Si+Sj+Sk)Eijk 
       *  Chiral Order Z */  
      r=0.0;for(i1=0;i1<nn;i1++){ z0 =zi(ib1,i1);
      /*  site 0..NN, direction 0..2, steps ..,-1,0,1,.. */
       sx =site_move(i1,1,0,0);     zx =zi(ib1,sx);
       sy =site_move(i1,0,1,0);     zy =zi(ib1,sy);
       sxy=site_move(sx,0,1,0);     zxy=zi(ib1,sxy);
       if(z0*zx <0)r+=zx *M_f2d(vec_e(v0,b2i(Flip(ib1,i1 ,sx )))); 
       if(zx*zxy<0)r+=zxy*M_f2d(vec_e(v0,b2i(Flip(ib1,sx ,sxy)))); 
       if(zxy*z0<0)r+=z0 *M_f2d(vec_e(v0,b2i(Flip(ib1,sxy,i1 )))); 
       if(z0*zxy<0)r-=zxy*M_f2d(vec_e(v0,b2i(Flip(ib1,i1 ,sxy)))); 
       if(zxy*zy<0)r-=zy *M_f2d(vec_e(v0,b2i(Flip(ib1,sxy,sy )))); 
       if(zy*z0 <0)r-=z0 *M_f2d(vec_e(v0,b2i(Flip(ib1,sy ,i1 ))));} 
      rz+=Norm(0.5*r/nn);}
    return rz;}

int wtatb(cplx *rta, cplx *rtb, tvec *v0){
    ULong i;tbase ib1;int i1,z0,sx,zx,sy,zy,sxy,zxy;cplx r,r1,r2;
    if (mpi_n>1) E0("missing mpi code on wtatb");
    (*rta)=(*rtb)=0.0;
    for(i=0l;i<n1;i++){ ib1=getl1(i);
      /* better ?? ######### <cfg|Op(i,j,k)*v0> ########################### */
      /* TA=<[1/N Sum(E(i,+x,+x+y)+E(+x+y,+y,i)+E(+y,i,+x)+E(+x,+x+y,+y)]^2>
       * TB=<[1/N Sum(E(i,+x,+x+y)+E(+x+y,+y,i)-E(+y,i,+x)-E(+x,+x+y,+y)]^2>
       * Eijk=Si(Sj x Sk)=1/(2i)Det([[Sim,Sip,Siz],[Sjm,Sjp,Sjz],[Skm,Skp,Skz]]) 
       * Chiral Order TA,TB */  
      r1=r2=0.0;for(i1=0;i1<nn;i1++){ z0 =zi(ib1,i1); 
       sx =site_move(i1,1,0,0);       zx =zi(ib1,sx);
       sy =site_move(i1,0,1,0);       zy =zi(ib1,sy);
       sxy=site_move(sx,0,1,0);       zxy=zi(ib1,sxy);  r=0.0;
       if(z0*zx <0)r+=zxy*zx*M_f2d(vec_e(v0,b2i(Flip(ib1,i1 ,sx ))));
       if(zx*zxy<0)r+=z0*zxy*M_f2d(vec_e(v0,b2i(Flip(ib1,sx ,sxy))));
       if(zxy*z0<0)r+=zx*z0 *M_f2d(vec_e(v0,b2i(Flip(ib1,sxy,i1 ))));
       if(zxy*zy<0)r+=z0*zy *M_f2d(vec_e(v0,b2i(Flip(ib1,sxy,sy ))));
       if(zy*z0 <0)r+=zxy*z0*M_f2d(vec_e(v0,b2i(Flip(ib1,sy ,i1 ))));
       if(z0*zxy<0)r+=zy*zxy*M_f2d(vec_e(v0,b2i(Flip(ib1,i1 ,sxy))));r1+=r;r2+=r;r=0.0;
       if(zy*z0 <0)r+=zx*z0 *M_f2d(vec_e(v0,b2i(Flip(ib1,sy ,i1 ))));
       if(z0*zx <0)r+=zy*zx *M_f2d(vec_e(v0,b2i(Flip(ib1,i1 ,sx ))));
       if(zx*zy <0)r+=z0*zy *M_f2d(vec_e(v0,b2i(Flip(ib1,sx ,sy ))));
       if(zx*zxy<0)r+=zy*zxy*M_f2d(vec_e(v0,b2i(Flip(ib1,sx ,sxy))));
       if(zxy*zy<0)r+=zx*zy *M_f2d(vec_e(v0,b2i(Flip(ib1,sxy,sy ))));
       if(zy*zx <0)r+=zxy*zx*M_f2d(vec_e(v0,b2i(Flip(ib1,sy ,sx ))));r1+=r;r2-=r;} 
      (*rta)+=Norm(0.25*r1/nn);
      (*rtb)+=Norm(0.25*r2/nn);}
    return 0;}
#endif	/* N1>1 */

